from __future__ import annotations

import textwrap
from pathlib import Path

import pytest
from rnzb import InvalidNzbError, Nzb

NZB_DIR = Path(__file__).parent.resolve() / "nzbs"

INVALID_NZB_ERROR_GROUPS_ELEMENT = "Invalid or missing 'groups' element within the 'file' element. Each 'file' element must contain at least one valid 'groups' element."
INVALID_NZB_ERROR_SEGMENTS_ELEMENT = "Invalid or missing 'segments' element within the 'file' element. Each 'file' element must contain at least one valid 'segments' element."
INVALID_NZB_ERROR_FILE_ELEMENT = "Invalid or missing 'file' element in the NZB document. The NZB document must contain at least one valid 'file' element, and each 'file' must have at least one valid 'groups' and 'segments' element."

invalid_xml = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
        <meta type="password">secret</meta>
        <meta type="tag">HD</meta>
        <meta type="category">TV</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
        <segments>
            <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
            <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
        </segments>
    </file>
"""

valid_xml_but_invalid_nzb = """\
<?xml version="1.0" encoding="iso-8859-1" ?>
<!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
<nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
    <head>
        <meta type="title">Your File!</meta>
    </head>
    <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
        <groups>
            <group>alt.binaries.newzbin</group>
            <group>alt.binaries.mojo</group>
        </groups>
    </file>
</nzb>"""


@pytest.mark.parametrize(
    "input_xml, expected_error",
    [
        pytest.param(
            invalid_xml,
            "^The NZB document is not valid XML and could not be parsed:",
            id="truncated_xml",
        ),
        pytest.param(valid_xml_but_invalid_nzb, INVALID_NZB_ERROR_SEGMENTS_ELEMENT, id="missing_segments"),
    ],
)
def test_parsing_invalid_nzb(input_xml: str, expected_error: str) -> None:
    with pytest.raises(InvalidNzbError, match=expected_error):
        Nzb.from_str(input_xml)


@pytest.mark.parametrize(
    "file_name, expected_error",
    (
        ("malformed_files.nzb", INVALID_NZB_ERROR_FILE_ELEMENT),
        ("malformed_files2.nzb", INVALID_NZB_ERROR_GROUPS_ELEMENT),
        ("malformed_groups.nzb", INVALID_NZB_ERROR_GROUPS_ELEMENT),
        ("malformed_segments.nzb", INVALID_NZB_ERROR_SEGMENTS_ELEMENT),
    ),
    ids=["malformed_files", "malformed_files2", "malformed_groups", "malformed_segments"],
)
def test_parser_exceptions(file_name: str, expected_error: str) -> None:
    with pytest.raises(InvalidNzbError, match=expected_error):
        Nzb.from_file(NZB_DIR / file_name)


def test_read_nzb_file(tmp_path: Path) -> None:
    with pytest.raises(InvalidNzbError, match="^Gzip decompression error for file"):
        Nzb.from_file(NZB_DIR / "invalid_gzipped_nzb.nzb.gz")

    tmp_file = tmp_path / "invalid_bytes.nzb"
    tmp_file.write_bytes(bytes([255]))

    with pytest.raises(InvalidNzbError, match="^I/O error while reading file"):
        Nzb.from_file(tmp_file)


def test_nzb_with_bad_file_date() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="not a date" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'date' in a 'file' element."):
        Nzb.from_str(nzb)


def test_nzb_with_missing_file_poster() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file date="1071674882" subject="Here's your file!  abc-mr2a.r01 (1/2)">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'poster' in a 'file' element."):
        Nzb.from_str(nzb)


def test_nzb_with_missing_file_subject() -> None:
    nzb = textwrap.dedent("""
    <?xml version="1.0" encoding="iso-8859-1" ?>
    <!DOCTYPE nzb PUBLIC "-//newzBin//DTD NZB 1.1//EN" "http://www.newzbin.com/DTD/nzb/nzb-1.1.dtd">
    <nzb xmlns="http://www.newzbin.com/DTD/2003/nzb">
        <head>
            <meta type="title">Your File!</meta>
            <meta type="password">secret</meta>
            <meta type="tag">HD</meta>
            <meta type="category">TV</meta>
        </head>
        <file poster="Joe Bloggs &lt;bloggs@nowhere.example&gt;" date="1071674882">
            <groups>
                <group>alt.binaries.newzbin</group>
                <group>alt.binaries.mojo</group>
            </groups>
            <segments>
                <segment bytes="102394" number="1">123456789abcdef@news.newzbin.com</segment>
                <segment bytes="4501" number="2">987654321fedbca@news.newzbin.com</segment>
            </segments>
        </file>
    </nzb>
    """).strip()
    with pytest.raises(InvalidNzbError, match=r"Invalid or missing required attribute 'subject' in a 'file' element."):
        Nzb.from_str(nzb)


def test_non_existent_file() -> None:
    with pytest.raises(FileNotFoundError, match="blahblah"):
        Nzb.from_file("blahblah")



from pathlib import Path

from rnzb import Nzb

NZB_DIR = Path(__file__).parent.resolve() / "nzbs"


def test_repr() -> None:
    nzb = Nzb.from_file(NZB_DIR / "spec_example.nzb")
    assert (
        repr(nzb.meta)
        == str(nzb.meta)
        == 'Meta(title="Your File!", passwords=("secret",), tags=("HD",), category="TV")'
    )
    assert repr(nzb.meta.passwords) == str(nzb.meta.passwords) == "('secret',)"
    assert repr(nzb.meta.tags) == str(nzb.meta.tags) == "('HD',)"
    assert (
        repr(nzb.file.segments[0])
        == str(nzb.file.segments[0])
        == 'Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com")'
    )
    assert (
        repr(nzb.file.segments)
        == str(nzb.file.segments)
        == '(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))'
    )
    assert (
        repr(nzb.file)
        == str(nzb.file)
        == 'File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com")))'
    )
    assert (
        repr(nzb.files)
        == str(nzb.files)
        == '(File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))),)'
    )
    assert (
        repr(nzb)
        == str(nzb)
        == 'Nzb(meta=Meta(title="Your File!", passwords=("secret",), tags=("HD",), category="TV"), files=(File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))),))'
    )


def test_repr_2() -> None:
    nzb = Nzb.from_file(NZB_DIR / "spec_example_with_multiple_meta.nzb")
    assert (
        repr(nzb.meta)
        == str(nzb.meta)
        == 'Meta(title="Your File!", passwords=(), tags=("HD", "1080p", "FLAC"), category="TV")'
    )
    assert repr(nzb.meta.passwords) == str(nzb.meta.passwords) == "()"
    assert repr(nzb.meta.tags) == str(nzb.meta.tags) == "('HD', '1080p', 'FLAC')"
    assert (
        repr(nzb.file.segments[0])
        == str(nzb.file.segments[0])
        == 'Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com")'
    )
    assert (
        repr(nzb.file.segments)
        == str(nzb.file.segments)
        == '(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))'
    )
    assert (
        repr(nzb.file)
        == str(nzb.file)
        == 'File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com")))'
    )
    assert (
        repr(nzb.files)
        == str(nzb.files)
        == '(File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))),)'
    )
    assert (
        repr(nzb)
        == str(nzb)
        == 'Nzb(meta=Meta(title="Your File!", passwords=(), tags=("HD", "1080p", "FLAC"), category="TV"), files=(File(poster="Joe Bloggs <bloggs@nowhere.example>", posted_at="2003-12-17T15:28:02+00:00", subject="Here\'s your file!  abc-mr2a.r01 (1/2)", groups=("alt.binaries.mojo", "alt.binaries.newzbin"), segments=(Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"), Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"))),))'
    )



from datetime import datetime, timezone

import pytest
from rnzb import File, Meta, Nzb, Segment


def date() -> datetime:
    return datetime(2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc)


def test_meta_constructor() -> None:
    meta1 = Meta(title="title", passwords=("password",), tags=("tag",), category="category")
    assert meta1.title == "title"
    assert meta1.passwords == ("password",)
    assert meta1.tags == ("tag",)
    assert meta1.category == "category"

    meta2 = Meta(title=None, passwords=["password"], tags=["tag"], category=None)
    assert meta2.title is None
    assert meta2.passwords == ("password",)
    assert meta2.tags == ("tag",)
    assert meta2.category is None

    meta3 = Meta()
    assert meta3.title is None
    assert meta3.passwords == ()
    assert meta3.tags == ()
    assert meta3.category is None

    with pytest.raises(TypeError):
        # Positional arguments are not allowed
        Meta("title", ("password",), ("tag",), "category")  # type: ignore[misc]


def test_segment_constructor() -> None:
    segment = Segment(size=1, number=1, message_id="message_id")
    assert segment.size == 1
    assert segment.number == 1
    assert segment.message_id == "message_id"

    with pytest.raises(TypeError):
        Segment()  # type: ignore[call-arg]

    with pytest.raises(TypeError):
        # Positional arguments are not allowed
        Segment(1, 1, "message_id")  # type: ignore[misc]


def test_file_constructor() -> None:
    segment = Segment(size=1, number=1, message_id="message_id")
    file = File(poster="poster", posted_at=date(), subject="subject", groups=["group"], segments=[segment])
    assert file.poster == "poster"
    assert file.posted_at == date()
    assert file.subject == "subject"
    assert file.groups == ("group",)
    assert file.segments == (segment,)

    with pytest.raises(TypeError):
        File()  # type: ignore[call-arg]

    with pytest.raises(TypeError):
        # Positional arguments are not allowed
        File("poster", date(), "subject", ["group"], [segment])  # type: ignore[misc]


def test_nzb_constructor() -> None:
    segment = Segment(size=1, number=1, message_id="message_id")
    file = File(poster="poster", posted_at=date(), subject="subject", groups=["group"], segments=[segment])
    nzb = Nzb(meta=Meta(), files=[file])
    assert nzb.meta == Meta()
    assert nzb.files == (file,)
    assert nzb.file == file

    with pytest.raises(TypeError):
        Nzb()  # type: ignore[call-arg]

    with pytest.raises(TypeError):
        # Positional arguments are not allowed
        Nzb(Meta(), [file])  # type: ignore[misc]



import sys
from collections.abc import Sequence
from datetime import datetime
from os import PathLike
from typing import final

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = ("File", "InvalidNzbError", "Meta", "Nzb", "Segment")

@final
class InvalidNzbError(Exception):
    """Raised when the NZB is invalid."""

@final
class Meta:
    """Optional creator-definable metadata for the contents of the NZB."""

    title: str | None
    """Title."""

    passwords: tuple[str, ...]
    """Passwords."""

    tags: tuple[str, ...]
    """Tags."""

    category: str | None
    """Category."""

    def __new__(
        cls,
        *,
        title: str | None = None,
        passwords: Sequence[str] = (),
        tags: Sequence[str] = (),
        category: str | None = None,
    ) -> Meta:
        """Create a new instance of Meta."""

    def __eq__(self, value: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Self: ...

@final
class Segment:
    """One part segment of a file."""

    size: int
    """Size of the segment."""
    number: int
    """Number of the segment."""
    message_id: str
    """Message ID of the segment."""

    def __new__(cls, *, size: int, number: int, message_id: str) -> Segment:
        """Create a new instance of Segment."""

    def __eq__(self, value: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Self: ...

@final
class File:
    """Represents a complete file, consisting of segments that make up a file."""

    poster: str
    """The poster of the file."""

    posted_at: datetime
    """The date and time when the file was posted, in UTC."""

    subject: str
    """The subject of the file."""

    groups: tuple[str, ...]
    """Groups that reference the file."""

    segments: tuple[Segment, ...]
    """Segments that make up the file."""

    def __new__(
        cls,
        *,
        poster: str,
        posted_at: datetime,
        subject: str,
        groups: Sequence[str],
        segments: Sequence[Segment],
    ) -> File:
        """Create a new instance of File."""

    def __eq__(self, value: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Self: ...
    @property
    def size(self) -> int:
        """Size of the file calculated from the sum of segment sizes."""

    @property
    def name(self) -> str | None:
        """
        Complete name of the file with it's extension extracted from the subject.
        May return `None` if it fails to extract the name.
        """

    @property
    def stem(self) -> str | None:
        """
        Base name of the file without it's extension extracted from the [`File.name`][rnzb.File.name].
        May return `None` if it fails to extract the stem.
        """

    @property
    def extension(self) -> str | None:
        """
        Extension of the file without the leading dot extracted from the [`File.name`][rnzb.File.name].
        May return `None` if it fails to extract the extension.
        """

    def has_extension(self, ext: str, /) -> bool:
        """
        Check if the file has the specified extension.

        This method ensures consistent extension comparison
        by normalizing the extension (removing any leading dot)
        and handling case-folding.

        Parameters
        ----------
        ext : str
            Extension to check for, with or without a leading dot (e.g., `.txt` or `txt`).

        Returns
        -------
        bool
            `True` if the file has the specified extension, `False` otherwise.

        Examples
        --------
        ```python
        >>> file.has_extension('.TXT')  # True for 'file.txt'
        True
        >>> file.has_extension('txt')   # Also True for 'file.txt'
        True
        ```

        """

    def is_par2(self) -> bool:
        """
        Return `True` if the file is a `.par2` file, `False` otherwise.
        """

    def is_rar(self) -> bool:
        """
        Return `True` if the file is a `.rar` file, `False` otherwise.
        """

    def is_obfuscated(self) -> bool:
        """
        Return `True` if the file is obfuscated, `False` otherwise.
        """

@final
class Nzb:
    """Represents a complete NZB file."""

    meta: Meta
    """Optional creator-definable metadata for the contents of the NZB."""

    files: tuple[File, ...]
    """File objects representing the files included in the NZB."""

    def __new__(cls, *, meta: Meta, files: Sequence[File]) -> Nzb:
        """Create a new instance of NZB."""

    def __eq__(self, value: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Self: ...
    @classmethod
    def from_str(cls, nzb: str, /) -> Nzb:
        """
        Parse the given string into an [`Nzb`][rnzb.Nzb].

        Parameters
        ----------
        nzb : str
            NZB string.

        Returns
        -------
        Nzb
            Object representing the parsed NZB file.

        Raises
        ------
        InvalidNzbError
            Raised if the NZB is invalid.

        """

    @classmethod
    def from_file(cls, nzb: str | PathLike[str], /) -> Nzb:
        """
        Parse the given file into an [`Nzb`][nzb.Nzb].
        Handles both regular and gzipped NZB files.

        Parameters
        ----------
        nzb : str | PathLike[str]
            Path to the NZB file.

        Returns
        -------
        Nzb
            Object representing the parsed NZB file.

        Raises
        ------
        FileNotFoundError
            Raised if the specified file doesn't exist.

        InvalidNzbError
            Raised if the NZB is invalid.

        """

    @classmethod
    def from_json(cls, json: str, /) -> Nzb:
        """
        Deserialize the given JSON string into an [`Nzb`][rnzb.Nzb].

        Parameters
        ----------
        json : str
            JSON string representing the NZB.

        Returns
        -------
        Nzb
            Object representing the parsed NZB file.

        Raises
        ------
        InvalidNzbError
            Raised if the NZB is invalid.

        """

    def to_json(self, *, pretty: bool = False) -> str:
        """
        Serialize the [`Nzb`][rnzb.Nzb] object into a JSON string.

        Parameters
        ----------
        pretty : bool, optional
            Whether to pretty format the JSON string.

        Returns
        -------
        str
            JSON string representing the NZB.

        """

    @property
    def file(self) -> File:
        """
        The main content file (episode, movie, etc) in the NZB.
        This is determined by finding the largest file in the NZB
        and may not always be accurate.
        """

    @property
    def size(self) -> int:
        """Total size of all the files in the NZB."""

    @property
    def filenames(self) -> tuple[str, ...]:
        """
        Tuple of unique file names across all the files in the NZB.
        May return an empty tuple if it fails to extract the name for every file.
        """

    @property
    def posters(self) -> tuple[str, ...]:
        """
        Tuple of unique posters across all the files in the NZB.
        """

    @property
    def groups(self) -> tuple[str, ...]:
        """
        Tuple of unique groups across all the files in the NZB.
        """

    @property
    def par2_files(self) -> tuple[File, ...]:
        """
        Tuple of par2 files in the NZB.
        """

    @property
    def par2_size(self) -> int:
        """
        Total size of all the `.par2` files.
        """

    @property
    def par2_percentage(self) -> float:
        """
        Percentage of the size of all the `.par2` files relative to the total size.
        """

    def has_extension(self, ext: str, /) -> bool:
        """
        Check if any file in the NZB has the specified extension.

        This method ensures consistent extension comparison
        by normalizing the extension (removing any leading dot)
        and handling case-folding.

        Parameters
        ----------
        ext : str
            Extension to check for, with or without a leading dot (e.g., `.txt` or `txt`).

        Returns
        -------
        bool
            `True` if any file in the NZB has the specified extension, `False` otherwise.
        ```

        """

    def has_rar(self) -> bool:
        """
        Return `True` if any file in the NZB is a `.rar` file, `False` otherwise.
        """

    def is_rar(self) -> bool:
        """
        Return `True` if all files in the NZB are `.rar` files, `False` otherwise.
        """

    def is_obfuscated(self) -> bool:
        """
        Return `True` if any file in the NZB is obfuscated, `False` otherwise.
        """

    def has_par2(self) -> bool:
        """
        Return `True` if there's at least one `.par2` file in the NZB, `False` otherwise.
        """



from __future__ import annotations

import datetime
import sys
from pathlib import Path

import pytest
from rnzb import File, Nzb, Segment

NZB_DIR = Path(__file__).parent.resolve() / "nzbs"


@pytest.mark.parametrize(
    "nzb_file",
    (
        NZB_DIR / "spec_example.nzb",
        NZB_DIR / "spec_example.nzb.gz",
    ),
    ids=lambda x: x.name,
)
def test_spec_example_nzb(nzb_file: Path) -> None:
    nzb = Nzb.from_file(nzb_file)
    assert nzb.meta.title == "Your File!"
    assert nzb.meta.passwords == ("secret",)
    assert nzb.meta.tags == ("HD",)
    assert nzb.meta.category == "TV"
    assert len(nzb.files) == 1
    assert nzb.is_rar() is True
    assert nzb.is_obfuscated() is True
    assert nzb.file.name == "abc-mr2a.r01"
    assert nzb.file.stem == "abc-mr2a"
    assert nzb.file.extension == "r01"
    assert nzb.size == 106895
    assert len(nzb.files[0].segments) == 2
    assert nzb.files[0].segments == (
        Segment(size=102394, number=1, message_id="123456789abcdef@news.newzbin.com"),
        Segment(size=4501, number=2, message_id="987654321fedbca@news.newzbin.com"),
    )

    assert nzb.files[0].groups == ("alt.binaries.mojo", "alt.binaries.newzbin")


@pytest.mark.parametrize(
    "nzb_file",
    (
        NZB_DIR / "big_buck_bunny.nzb",
        NZB_DIR / "big_buck_bunny.nzb.gz",
    ),
    ids=lambda x: x.name,
)
def test_big_buck_bunny(nzb_file: Path) -> None:
    nzb = Nzb.from_file(nzb_file)

    assert nzb.meta.title is None
    assert nzb.meta.passwords == ()
    assert nzb.meta.tags == ()
    assert nzb.meta.category is None
    assert len(nzb.files) == 5
    assert nzb.is_rar() is False
    assert nzb.is_obfuscated() is False
    assert nzb.has_par2() is True
    assert nzb.file.name == "Big Buck Bunny - S01E01.mkv"
    assert nzb.file.stem == "Big Buck Bunny - S01E01"
    assert nzb.file.extension == "mkv"
    assert nzb.has_extension(".mkv") is True
    assert nzb.has_extension("mkv") is True
    assert nzb.has_extension("blah") is False
    assert nzb.size == 22704889
    assert [file.subject for file in nzb.files] == [
        '[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        '[2/5] - "Big Buck Bunny - S01E01.mkv.par2" yEnc (1/1) 920',
        '[3/5] - "Big Buck Bunny - S01E01.mkv.vol00+01.par2" yEnc (1/2) 717788',
        '[4/5] - "Big Buck Bunny - S01E01.mkv.vol01+02.par2" yEnc (1/3) 1434656',
        '[5/5] - "Big Buck Bunny - S01E01.mkv.vol03+04.par2" yEnc (1/5) 2869192',
    ]
    assert [file.subject for file in nzb.par2_files] == [
        '[2/5] - "Big Buck Bunny - S01E01.mkv.par2" yEnc (1/1) 920',
        '[3/5] - "Big Buck Bunny - S01E01.mkv.vol00+01.par2" yEnc (1/2) 717788',
        '[4/5] - "Big Buck Bunny - S01E01.mkv.vol01+02.par2" yEnc (1/3) 1434656',
        '[5/5] - "Big Buck Bunny - S01E01.mkv.vol03+04.par2" yEnc (1/5) 2869192',
    ]
    assert nzb.filenames == (
        "Big Buck Bunny - S01E01.mkv",
        "Big Buck Bunny - S01E01.mkv.par2",
        "Big Buck Bunny - S01E01.mkv.vol00+01.par2",
        "Big Buck Bunny - S01E01.mkv.vol01+02.par2",
        "Big Buck Bunny - S01E01.mkv.vol03+04.par2",
    )
    assert nzb.posters == ("John <nzb@nowhere.example>",)
    assert nzb.groups == ("alt.binaries.boneless",)
    assert nzb.par2_size == 5183128
    assert nzb.par2_percentage == pytest.approx(22, 1.0)
    assert nzb.file == File(
        poster="John <nzb@nowhere.example>",
        posted_at=datetime.datetime(2024, 1, 28, 11, 18, 28, tzinfo=datetime.timezone.utc),
        subject='[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        groups=("alt.binaries.boneless",),
        segments=(
            Segment(size=739067, number=1, message_id="9cacde4c986547369becbf97003fb2c5-9483514693959@example"),
            Segment(size=739549, number=2, message_id="70a3a038ce324e618e2751e063d6a036-7285710986748@example"),
            Segment(size=739728, number=3, message_id="a209875cefd44440aa91590508b48f5b-4625756912881@example"),
            Segment(size=739664, number=4, message_id="44057720ed4e45e4bce21d53249d03f8-8250738040266@example"),
            Segment(size=739645, number=5, message_id="cfc13d14583c484483aa49ac420bad27-9491395432062@example"),
            Segment(size=739538, number=6, message_id="5e90857531be401e9d0b632221fe2fb7-9854527985639@example"),
            Segment(size=739708, number=7, message_id="c33a2bba79494840a09d750b19d3b287-2550637855678@example"),
            Segment(size=739490, number=8, message_id="38006019d94f4ecc8f19c389c00f1ebe-7841585708380@example"),
            Segment(size=739667, number=9, message_id="b75a2425bef24fd5affb00dc3db789f6-7051027232703@example"),
            Segment(size=739540, number=10, message_id="79a027e3bfde458ea2bd0db1632fc84e-7270120407913@example"),
            Segment(size=739657, number=11, message_id="fb2bd74e1257487a9240ef0cf81765cc-7147741101314@example"),
            Segment(size=739647, number=12, message_id="d39ca8be78c34e3fa6f3211f1b397b3a-4725950858191@example"),
            Segment(size=739668, number=13, message_id="a4c15599055848dda1eff3b6b406fa78-8111735210252@example"),
            Segment(size=739721, number=14, message_id="2f1cec363ed24584b4127af86ac312ad-7204153818612@example"),
            Segment(size=739740, number=15, message_id="30ff3514896543a8ac91ec80346a5d40-9134304686352@example"),
            Segment(size=739538, number=16, message_id="1f75cfa20d884b5b972cfd2e9ebef249-8919850122587@example"),
            Segment(size=739646, number=17, message_id="8e22b0f973de4393a0a30ab094565316-6722799721412@example"),
            Segment(size=739610, number=18, message_id="faddf83650cc4de1a8bee68cffca40a1-5979589815618@example"),
            Segment(size=739514, number=19, message_id="6b8c23e43d4240da812b547babdc0423-6409257710918@example"),
            Segment(size=739920, number=20, message_id="802bd0dcef134ac690044e0a09fece60-8492061912475@example"),
            Segment(size=739634, number=21, message_id="efc4b3966a1f4b7787677e9e9a214727-5444471572012@example"),
            Segment(size=739691, number=22, message_id="247efca709114fd181bcaef0f487925f-4076317880026@example"),
            Segment(size=739638, number=23, message_id="665d9fc5edba4faca68ae835b702b4c7-9814601723860@example"),
            Segment(size=510541, number=24, message_id="962fddf3e07444988731b52aeaa9b2aa-1283919353788@example"),
        ),
    )


@pytest.mark.parametrize(
    "nzb_file",
    (
        NZB_DIR / "valid_nzb_with_bad_segments.nzb",
        NZB_DIR / "valid_nzb_with_bad_segments.nzb.gz",
    ),
    ids=lambda x: x.name,
)
def test_valid_nzb_with_one_missing_segment(nzb_file: Path) -> None:
    nzb = Nzb.from_file(nzb_file)

    assert nzb.has_extension(".mkv") is True
    assert nzb.has_extension("mkv") is True
    assert nzb.has_extension("blah") is False

    assert [file.subject for file in nzb.files] == [
        '[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        '[2/5] - "Big Buck Bunny - S01E01.mkv.par2" yEnc (1/1) 920',
        '[3/5] - "Big Buck Bunny - S01E01.mkv.vol00+01.par2" yEnc (1/2) 717788',
        '[4/5] - "Big Buck Bunny - S01E01.mkv.vol01+02.par2" yEnc (1/3) 1434656',
        '[5/5] - "Big Buck Bunny - S01E01.mkv.vol03+04.par2" yEnc (1/5) 2869192',
    ]
    assert [file.subject for file in nzb.par2_files] == [
        '[2/5] - "Big Buck Bunny - S01E01.mkv.par2" yEnc (1/1) 920',
        '[3/5] - "Big Buck Bunny - S01E01.mkv.vol00+01.par2" yEnc (1/2) 717788',
        '[4/5] - "Big Buck Bunny - S01E01.mkv.vol01+02.par2" yEnc (1/3) 1434656',
        '[5/5] - "Big Buck Bunny - S01E01.mkv.vol03+04.par2" yEnc (1/5) 2869192',
    ]
    assert nzb.filenames == (
        "Big Buck Bunny - S01E01.mkv",
        "Big Buck Bunny - S01E01.mkv.par2",
        "Big Buck Bunny - S01E01.mkv.vol00+01.par2",
        "Big Buck Bunny - S01E01.mkv.vol01+02.par2",
        "Big Buck Bunny - S01E01.mkv.vol03+04.par2",
    )

    assert nzb.file == File(
        poster="John <nzb@nowhere.example>",
        posted_at=datetime.datetime(2024, 1, 28, 11, 18, 28, tzinfo=datetime.timezone.utc),
        subject='[1/5] - "Big Buck Bunny - S01E01.mkv" yEnc (1/24) 16981056',
        groups=("alt.binaries.boneless",),
        segments=(
            Segment(size=739067, number=1, message_id="9cacde4c986547369becbf97003fb2c5-9483514693959@example"),
            Segment(size=739549, number=2, message_id="70a3a038ce324e618e2751e063d6a036-7285710986748@example"),
            Segment(size=739728, number=3, message_id="a209875cefd44440aa91590508b48f5b-4625756912881@example"),
            Segment(size=739664, number=4, message_id="44057720ed4e45e4bce21d53249d03f8-8250738040266@example"),
            Segment(size=739645, number=5, message_id="cfc13d14583c484483aa49ac420bad27-9491395432062@example"),
            Segment(size=739538, number=6, message_id="5e90857531be401e9d0b632221fe2fb7-9854527985639@example"),
            Segment(size=739708, number=7, message_id="c33a2bba79494840a09d750b19d3b287-2550637855678@example"),
            Segment(size=739490, number=8, message_id="38006019d94f4ecc8f19c389c00f1ebe-7841585708380@example"),
            Segment(size=739667, number=9, message_id="b75a2425bef24fd5affb00dc3db789f6-7051027232703@example"),
            Segment(size=739540, number=10, message_id="79a027e3bfde458ea2bd0db1632fc84e-7270120407913@example"),
            # 11-13 segments are missing here
            Segment(size=739721, number=14, message_id="2f1cec363ed24584b4127af86ac312ad-7204153818612@example"),
            Segment(size=739740, number=15, message_id="30ff3514896543a8ac91ec80346a5d40-9134304686352@example"),
            Segment(size=739538, number=16, message_id="1f75cfa20d884b5b972cfd2e9ebef249-8919850122587@example"),
            Segment(size=739646, number=17, message_id="8e22b0f973de4393a0a30ab094565316-6722799721412@example"),
            Segment(size=739610, number=18, message_id="faddf83650cc4de1a8bee68cffca40a1-5979589815618@example"),
            Segment(size=739514, number=19, message_id="6b8c23e43d4240da812b547babdc0423-6409257710918@example"),
            Segment(size=739920, number=20, message_id="802bd0dcef134ac690044e0a09fece60-8492061912475@example"),
            Segment(size=739634, number=21, message_id="efc4b3966a1f4b7787677e9e9a214727-5444471572012@example"),
            Segment(size=739691, number=22, message_id="247efca709114fd181bcaef0f487925f-4076317880026@example"),
            Segment(size=739638, number=23, message_id="665d9fc5edba4faca68ae835b702b4c7-9814601723860@example"),
            Segment(size=510541, number=24, message_id="962fddf3e07444988731b52aeaa9b2aa-1283919353788@example"),
        ),
    )


def test_bad_subject() -> None:
    nzb = Nzb.from_file(NZB_DIR / "bad_subject.nzb")
    assert nzb.files[0].name is None
    assert nzb.files[0].stem is None
    assert nzb.files[0].extension is None
    assert nzb.files[0].has_extension("mkv") is False
    assert nzb.files[0].is_par2() is False
    assert nzb.files[0].is_rar() is False
    assert nzb.is_rar() is False
    assert nzb.has_par2() is False
    assert nzb.is_obfuscated() is True


def test_non_standard_meta() -> None:
    nzb = Nzb.from_file(NZB_DIR / "non_standard_meta.nzb")
    assert nzb.meta.title is None
    assert nzb.meta.passwords == ()
    assert nzb.meta.tags == ()
    assert nzb.meta.category is None


def test_single_rar_nzb() -> None:
    nzb = Nzb.from_file(NZB_DIR / "one_rar_file.nzb")
    assert nzb.has_rar() is True
    assert nzb.is_rar() is False
    assert nzb.has_par2() is False


def test_multi_rar_nzb() -> None:
    nzb = Nzb.from_file(NZB_DIR / "multi_rar.nzb")
    assert nzb.has_rar() is True
    assert nzb.is_rar() is True
    assert nzb.has_par2() is False


@pytest.mark.skipif(
    (
        sys.version_info < (3, 10)  # nzb doesn't support < 3.10
        or sys.version_info >= (3, 13)  # nzb depends on msgspec, which does not publish wheels for 3.13 free-threaded.
    )
    or sys.implementation.name != "cpython",  # nzb depends on msgspec, which does not publish wheels for pypy.
    reason="This test requires 'nzb' which depends on 'msgspec' and 'msgspec' does not publish wheels for pypy or 3.13t.",
)
@pytest.mark.parametrize(
    "nzb_file",
    [
        NZB_DIR / "spec_example.nzb",
        NZB_DIR / "spec_example.nzb.gz",
        NZB_DIR / "big_buck_bunny.nzb",
        NZB_DIR / "big_buck_bunny.nzb.gz",
        NZB_DIR / "bad_subject.nzb",
        NZB_DIR / "non_standard_meta.nzb",
        NZB_DIR / "one_rar_file.nzb",
        NZB_DIR / "multi_rar.nzb",
        NZB_DIR / "valid_nzb_with_bad_segments.nzb",
        NZB_DIR / "valid_nzb_with_bad_segments.nzb.gz",
    ],
    ids=lambda x: x.name,
)
def test_json_roundtrip(nzb_file: Path) -> None:
    from nzb import Nzb as PurePythonNzb

    original = PurePythonNzb.from_file(nzb_file)
    original_rnzb = Nzb.from_file(nzb_file)

    serialized = original.to_json()
    serialized_rnzb = original_rnzb.to_json()

    assert serialized == serialized_rnzb

    deserialized = PurePythonNzb.from_json(serialized)
    deserialized_rnzb = Nzb.from_json(serialized_rnzb)

    assert original == deserialized
    assert original_rnzb == deserialized_rnzb

    assert deserialized.to_json(pretty=True) == deserialized_rnzb.to_json(pretty=True)



import os
import shutil

import nox

nox.needs_version = ">=2024.10.9"
nox.options.default_venv_backend = "uv"

PYTHON_VERSIONS = ("3.9", "3.10", "3.11", "3.12", "3.13", "pypy3.9", "pypy3.10")


def cargo(session: nox.Session, *args: str) -> None:
    session.run("cargo", *args, external=True)


def install(session: nox.Session) -> None:
    session.run_install(
        "uv",
        "sync",
        "--only-group",
        "dev",
        "--locked",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )
    session.run("maturin", "develop", "--uv")


@nox.session(venv_backend=None, default=False)
def clean(session: nox.Session) -> None:
    cargo(session, "clean")

    paths = (
        "./.mypy_cache",
        "./.pytest_cache",
        "./.ruff_cache",
        "./dist",
        "./tests/__pycache__",
        "./__pycache__",
        "./.nox",
    )

    for p in paths:
        try:
            shutil.rmtree(p)
        except FileNotFoundError:
            pass


@nox.session
def lint(session: nox.Session) -> None:
    install(session)

    if os.getenv("CI"):
        # Do not modify files in CI, simply fail.
        cargo(session, "fmt", "--check")
        cargo(session, "clippy")
        session.run("ruff", "check", ".")
        session.run("ruff", "format", ".", "--check")
    else:
        # Fix any fixable errors if running locally.
        cargo(session, "fmt")
        cargo(session, "clippy", "--fix", "--lib", "-p", "rnzb", "--allow-dirty")
        session.run("ruff", "check", ".", "--fix")
        session.run("ruff", "format", ".")


@nox.session(python=PYTHON_VERSIONS)
def tests(session: nox.Session) -> None:
    install(session)
    session.run("pytest", "-vv", *session.posargs)
